package com.contextspace;

import static org.junit.Assert.*;

import java.awt.print.Printable;
import java.io.File;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;
import org.springframework.context.support.ClassPathXmlApplicationContext;

import com.contextspace.objectdb_example.dao.IAttributeNameDAO;
import com.contextspace.objectdb_example.dao.IHolderDAO;
import com.contextspace.objectdb_example.dao.IOtherObjectDAO;
import com.contextspace.objectdb_example.model.AttributeName;
import com.contextspace.objectdb_example.model.OtherObject;
import com.contextspace.objectdb_example.model.holder.Holder;

public class ExampleTest {

    private static ClassPathXmlApplicationContext context = null;
    
    // attribute
    private static IAttributeNameDAO attrNameDao;
    private static IHolderDAO holderDao;
    private static IOtherObjectDAO otherObjectDao;
    
    @BeforeClass
    public static void setUpBeforeClass() throws Exception {
        context = new ClassPathXmlApplicationContext(new String[] { "applicationContext-test.xml" });
        
        attrNameDao = context.getBean(IAttributeNameDAO.class);
        holderDao = context.getBean(IHolderDAO.class);
        otherObjectDao = context.getBean(IOtherObjectDAO.class);
    }
    
    @AfterClass
    public static void tearDownAfterClass() throws Exception {
        if(context != null) {
            context.close();
        }      
        
        File f1 = new File("objectdb"+File.separator+"db");
        deleteDir(f1);
    }
    
    
    @Test
    public void test() throws Exception {
        
        assertEquals(0, attrNameDao.count());
        
        AttributeName an1 = new AttributeName("an1", "an1 eng");
        AttributeName an2 = new AttributeName("an2", "an2 eng");
        an2.setDisplayValue("fr", "an2 fr");
        AttributeName an3 = new AttributeName("an3", "an3 eng");
        an3.setDisplayValue("fr", "an3 fr");
        
        an1 = attrNameDao.create(an1);
        an2 = attrNameDao.create(an2);
        an3 = attrNameDao.create(an3);
        
        assertEquals(3, attrNameDao.count());
        
        assertEquals("<attributename name=\"an1\" values=\"en=an1 eng\" id=\"1\"/>", attrNameDao.find(an1.getId()).toString());
        assertEquals("<attributename name=\"an2\" values=\"fr=an2 fr,en=an2 eng\" id=\"2\"/>", attrNameDao.find(an2.getId()).toString());
        assertEquals("<attributename name=\"an3\" values=\"fr=an3 fr,en=an3 eng\" id=\"3\"/>", attrNameDao.find(an3.getId()).toString());
        
        
        Holder holder = new Holder();
        holder = holderDao.create(holder);
        assertEquals("<holder id=\"4\">null</holder>", holderDao.find(holder.getId()).toString());
        
        
        Holder holder2 = new Holder();
        holder2.addName(an1);
        
        holder2 = holderDao.create(holder2);
        
        //get it back using a full select... THIS WORKS
        assertEquals("<holder id=\"5\">null<list><attributename name=\"an1\" values=\"en=an1 eng\" id=\"1\"/></list></holder>", holderDao.lookupById(holder2.getId()).toString());
        
        //get it back using a find... THIS FAILS
        assertEquals("<holder id=\"5\">null<list><attributename name=\"an1\" values=\"en=an1 eng\" id=\"1\"/></list></holder>", holderDao.find(holder2.getId()).toString());

        holderDao.delete(holder);
        holderDao.delete(holder2);
        assertEquals(0, holderDao.count());
        
        attrNameDao.delete(an1);
        attrNameDao.delete(an2);
        attrNameDao.delete(an3);
        
        assertEquals(0, attrNameDao.count());
    }

    
    @Test
    public void test2() throws Exception {
        
        AttributeName an1 = new AttributeName("holdername", "Holder Name");
        AttributeName an2 = new AttributeName("listname1", "List Name1");
        AttributeName an3 = new AttributeName("listname2", "List Name2");
        AttributeName an4 = new AttributeName("listname3", "List Name3");
        
        AttributeName an5 = new AttributeName("mapentry1", "Map Entry1");
        AttributeName an6 = new AttributeName("mapentry2", "Map Entry2");
        
        an1 = attrNameDao.create(an1);
        an2 = attrNameDao.create(an2);
        an3 = attrNameDao.create(an3);
        an4 = attrNameDao.create(an4);
        an5 = attrNameDao.create(an5);
        an6 = attrNameDao.create(an6);
        
        OtherObject oobj1 = new OtherObject("mapentry1 otherobject1");
        OtherObject oobj2 = new OtherObject("mapentry1 otherobject2");
        OtherObject oobj3 = new OtherObject("mapentry2 otherobject1");
        OtherObject oobj4 = new OtherObject("mapentry2 otherobject2");
        
        oobj1 = otherObjectDao.create(oobj1);
        oobj2 = otherObjectDao.create(oobj2);
        oobj3 = otherObjectDao.create(oobj3);
        oobj4 = otherObjectDao.create(oobj4);
        
        List<OtherObject> mapentry1OtherObjects = new ArrayList<OtherObject>();
        mapentry1OtherObjects.add(oobj1);
        mapentry1OtherObjects.add(oobj2);
        
        List<OtherObject> mapentry2OtherObjects = new ArrayList<OtherObject>();
        mapentry2OtherObjects.add(oobj3);
        mapentry2OtherObjects.add(oobj4);
        
        
        assertEquals(0, holderDao.count());
        
        Holder holder = new Holder();
        holder.setName(an1);
        holder.addName(an2);
        holder.addName(an3);
        holder.addName(an4);
        holder.addMapEntry(an5, mapentry1OtherObjects);
        holder.addMapEntry(an6, mapentry2OtherObjects);
        
        System.out.println("HashCode for an5: " + an5.hashCode());
        System.out.println("HashCode for an6: " + an6.hashCode());
        System.out.println("an5 == an6: " + an5.equals(an6));
        assertFalse(an5.hashCode() == an6.hashCode());
        assertFalse(an5.equals(an6));
        
        System.out.println("BEFORE: " + printPrettyXml(holder.toString()) + "\n\n");
        
        holder = holderDao.create(holder);     
        System.out.println("STRAIGHT AFTER CREATE: " + printPrettyXml(holder.toString()) + "\n\n");
        holderDao.flush();
        System.out.println("STRAIGHT AFTER FLUSH: " + printPrettyXml(holder.toString()) + "\n\n");
        holder = holderDao.find(holder.getId());
      
        System.out.println("AFTER: " + printPrettyXml(holder.toString()));
        assertEquals(2, holder.getMapOfOtherObjects().size());
        
        //OK, so its all happy to here, now what happens if I do an update?
        AttributeName an7 = new AttributeName("mapentry3", "Map Entry3");
        an7 = attrNameDao.create(an7);
        
        OtherObject oobj5 = new OtherObject("mapentry3 otherobject1");
        OtherObject oobj6 = new OtherObject("mapentry3 otherobject2");
        
        oobj5 = otherObjectDao.create(oobj5);
        oobj6 = otherObjectDao.create(oobj6);
        
        List<OtherObject> mapentry3OtherObjects = new ArrayList<OtherObject>();
        mapentry3OtherObjects.add(oobj5);
        mapentry3OtherObjects.add(oobj6);
        
        holder.addMapEntry(an7, mapentry3OtherObjects);
        
        System.out.println("BEFORE: " + printPrettyXml(holder.toString()) + "\n\n");
        assertEquals(3, holder.getMapOfOtherObjects().size());
        
        holder = holderDao.update(holder);
        
        holderDao.flush();
        
        holder = holderDao.find(holder.getId());
        //System.out.println("AFTER: " + printPrettyXml(holder.toString()));
        assertEquals(3, holder.getMapOfOtherObjects().size());
        
        //ok, so it happy to here, no issues doing an update.
        //what if we add a nested holder?
        
        AttributeName an8 = new AttributeName("holdername2", "Holder Name2");
        an8 = attrNameDao.create(an8);
        
        AttributeName an9 = new AttributeName("mapentry4", "Map Entry4");
        an9 = attrNameDao.create(an9);
        
        OtherObject oobj7 = new OtherObject("mapentry4 otherobject1");
        OtherObject oobj8 = new OtherObject("mapentry4 otherobject2");
        
        oobj7 = otherObjectDao.create(oobj7);
        oobj8 = otherObjectDao.create(oobj8);
        
        List<OtherObject> mapentry4OtherObjects = new ArrayList<OtherObject>();
        mapentry4OtherObjects.add(oobj7);
        mapentry4OtherObjects.add(oobj8);
        
        Holder holder2 = new Holder();
        holder2.setName(an8);
        holder2.addMapEntry(an9, mapentry4OtherObjects);
        
        holder.setHolder(holder2);
        
        //System.out.println("BEFORE: " + printPrettyXml(holder.toString()) + "\n\n");
        
        holder = holderDao.update(holder);
        
        holderDao.flush();
        
        holder = holderDao.find(holder.getId());
        //System.out.println("AFTER: " + printPrettyXml(holder.toString()));
        
        //ok, all good to here, what happens if we now add a map to the nested holder?
        holder2 = holder.getHolder();
     
        an9 = attrNameDao.find(an9.getId());
        System.out.println("an9: " + an9);
        
        Map<AttributeName, List<OtherObject>>testMap = holder2.getMapOfOtherObjects();
        assertNotNull(testMap);
        System.out.println(testMap);
        
        for(Iterator<AttributeName> i = testMap.keySet().iterator(); i.hasNext();) {
            AttributeName attrName = i.next();
            System.out.println("Looking at mapkey: " + attrName + ". Does it equal an9?: " + attrName.equals(an9));
        }
        
        System.out.println("Does testMap contain an9: " + testMap.containsKey(an9));
        List<OtherObject> testList = testMap.get(an9);
        assertNotNull(testList);
        assertTrue(!testList.isEmpty());
        
        OtherObject otherObject = testList.get(0);
        assertNotNull(otherObject);
        otherObject.addValue("key1", "value1");
        
        System.out.println("BEFORE: " + printPrettyXml(holder.toString()) + "\n\n");
        
        holder = holderDao.update(holder);
        
        holderDao.flush();
        
        holder = holderDao.find(holder.getId());
        System.out.println("AFTER: " + printPrettyXml(holder.toString()));
        
        holder2 = holder.getHolder();
        an9 = attrNameDao.find(an9.getId());
        testMap = holder2.getMapOfOtherObjects();
        testList = testMap.get(an9);
        assertNotNull(testList);
        assertTrue(!testList.isEmpty());
        OtherObject retrievedObject = testList.get(0);
        assertEquals(1, retrievedObject.getSomeValues().size());
        
    }
    
    private static boolean deleteDir(File dir) {
        if (dir.isDirectory()) {
            String[] children = dir.list();
            for (int i=0; i<children.length; i++) {
                boolean success = deleteDir(new File(dir, children[i]));
                if (!success) {
                    return false;
                }
            }
        }
    
        // The directory is now empty so delete it
        return dir.delete();
    } 
    
    public static String printPrettyXml(String input) {
        return printPrettyXml(input, 4);
    }
    
    public static String printPrettyXml(String input, int indent) {
        try {
            Source xmlInput = new StreamSource(new StringReader(input));
            StringWriter stringWriter = new StringWriter();
            StreamResult xmlOutput = new StreamResult(stringWriter);
            TransformerFactory transformerFactory = TransformerFactory.newInstance();
            Transformer transformer = transformerFactory.newTransformer();
            
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", String.valueOf(indent));
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.transform(xmlInput, xmlOutput);
            return xmlOutput.getWriter().toString();
        }
        catch(TransformerException e) {
            return "Exception thrown during transform of xml: " + e.getMessage();
        }
    }
}
